﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;

namespace TestProject
{
    /// <summary>
    /// GIF 파트
    /// </summary>
    public abstract class GIFPart
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region GIF 파트 데이터 - GIFPartData

        /// <summary>
        /// GIF 파트 데이터
        /// </summary>
        public abstract class GIFPartData : GIFPart
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Field
            ////////////////////////////////////////////////////////////////////////////////////////// Protected

            #region Field

            /// <summary>
            /// 데이터 바이트 배열
            /// </summary>
            protected byte[] dataByteArray = null;

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Field
            ////////////////////////////////////////////////////////////////////////////////////////// Protected

            #region 길이 - Length

            /// <summary>
            /// 길이
            /// </summary>
            public override int Length
            {
                get
                {
                    return this.dataByteArray == null ? 0 : this.dataByteArray.Length;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - GIFPartData()

            /// <summary>
            /// 생성자
            /// </summary>
            public GIFPartData()
            {
            }

            #endregion
            #region 생성자 - GIFPartData(dataByteArray, index, length)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray"></param>
            /// <param name="index"></param>
            /// <param name="length"></param>
            public GIFPartData(ref byte[] dataByteArray, ref int index, int length)
            {
                this.dataByteArray = new byte[length];

                Array.Copy(dataByteArray, index, this.dataByteArray, 0, length);

                index += length;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 데이터 구하기 - GetData()

            /// <summary>
            /// 데이터 구하기
            /// </summary>
            /// <returns>데이터</returns>
            public override byte[] GetDataByteArray()
            {
                return this.dataByteArray;
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Protected

            #region 값 구하기 - GetValue(index)

            /// <summary>
            /// 값 구하기
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <returns>값</returns>
            protected ushort GetValue(int index)
            {
                return Convert.ToUInt16(this.dataByteArray[index] | (this.dataByteArray[index + 1] << 8));
            }

            #endregion
            #region 값 설정하기 - SetValue(index, value)

            /// <summary>
            /// 값 설정하기
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <param name="value">값</param>
            protected void SetValue(int index, ushort value)
            {
                this.dataByteArray[index] = Convert.ToByte(value & 0xFF);

                this.dataByteArray[index + 1] = Convert.ToByte((value & 0xFF00) >> 8);
            }

            #endregion
            #region 문자열 구하기 - GetString(index, length)

            /// <summary>
            /// 문자열 구하기
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <param name="length">길이</param>
            /// <returns>문자열</returns>
            protected string GetString(int index, int length)
            {
                return Encoding.ASCII.GetString(this.dataByteArray, index, length);
            }

            #endregion
            #region 문자열 설정하기 - SetString(index, length, value)

            /// <summary>
            /// 문자열 설정하기
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <param name="length">길이</param>
            /// <param name="value">값</param>
            protected void SetString(int index, int length, string value)
            {
                byte[] bufferByteArray = Encoding.ASCII.GetBytes(value);

                if(bufferByteArray.Length != length)
                {
                    throw new Exception("The strings length must be " + length.ToString() + "!");
                }

                for(int i = 0; i < length; i++)
                {
                    this.dataByteArray[index + i] = bufferByteArray[i];
                }
            }

            #endregion
        }

        #endregion
        #region GIF 파트 리스트 -GIFPartList

        /// <summary>
        /// GIF 파트 리스트
        /// </summary>
        public abstract class GIFPartList : GIFPart, IList<GIFPart>
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 인덱서 - this[index]

            /// <summary>
            /// 인덱서
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <returns>GIF 파트</returns>
            public GIFPart this[int index]
            {
                get
                {
                    return this.partList[index];
                }
                set
                {
                    this.partList[index] = value;
                }
            }

            #endregion
            #region 카운트 - Count

            /// <summary>
            /// 카운트
            /// </summary>
            public int Count
            {
                get
                {
                    return this.partList.Count;
                }
            }

            #endregion
            #region 읽기 전용 여부 - IsReadOnly

            /// <summary>
            /// 읽기 전용 여부
            /// </summary>
            public bool IsReadOnly
            {
                get
                {
                    return false;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 추가하기 - Add(part)

            /// <summary>
            /// 추가하기
            /// </summary>
            /// <param name="part">GIF 파트</param>
            public void Add(GIFPart part)
            {
                this.partList.Add(part);
            }

            #endregion
            #region 지우기 - Clear()

            /// <summary>
            /// 지우기
            /// </summary>
            public void Clear()
            {
                this.partList.Clear();
            }

            #endregion
            #region 포함 여부 구하기 - Contains(part)

            /// <summary>
            /// 포함 여부 구하기
            /// </summary>
            /// <param name="part">GIF 파트</param>
            /// <returns>포함 여부</returns>
            public bool Contains(GIFPart part)
            {
                return this.partList.Contains(part);
            }

            #endregion
            #region 복사하기 - CopyTo(partArray, arrayIndex)

            /// <summary>
            /// 복사하기
            /// </summary>
            /// <param name="partArray">GIF 파트 배열</param>
            /// <param name="arrayIndex">배열 인덱스</param>
            public void CopyTo(GIFPart[] partArray, int arrayIndex)
            {
                this.partList.CopyTo(partArray, arrayIndex);
            }

            #endregion
            #region 열거자 구하기 - GetEnumerator()

            /// <summary>
            /// 열거자 구하기
            /// </summary>
            /// <returns>열거자</returns>
            public IEnumerator<GIFPart> GetEnumerator()
            {
                return this.partList.GetEnumerator();
            }

            #endregion
            #region 인덱스 구하기 - IndexOf(part)

            /// <summary>
            /// 인덱스 구하기
            /// </summary>
            /// <param name="part">GIF 파트</param>
            /// <returns>인덱스</returns>
            public int IndexOf(GIFPart part)
            {
                return this.partList.IndexOf(part);
            }

            #endregion
            #region 삽입하기 - Insert(index, part)

            /// <summary>
            /// 삽입하기
            /// </summary>
            /// <param name="index"></param>
            /// <param name="part"></param>
            public void Insert(int index, GIFPart part)
            {
                this.partList.Insert(index, part);
            }

            #endregion
            #region 제거하기 - Remove(part)

            /// <summary>
            /// 제거하기
            /// </summary>
            /// <param name="part">GIF 파트</param>
            /// <returns>처리 결과</returns>
            public bool Remove(GIFPart part)
            {
                return this.partList.Remove(part);
            }

            #endregion
            #region 제거하기 - RemoveAt(index)

            /// <summary>
            /// 제거하기
            /// </summary>
            /// <param name="index">인덱스</param>
            public void RemoveAt(int index)
            {
                this.partList.RemoveAt(index);
            }

            #endregion
            #region 열거자 구하기 - IEnumerable.GetEnumerator()

            /// <summary>
            /// 열거자 구하기
            /// </summary>
            /// <returns></returns>
            IEnumerator IEnumerable.GetEnumerator()
            {
                return (this.partList as IEnumerable).GetEnumerator();
            }

            #endregion
        }

        #endregion
        #region GIF 파서 예외 - GIFParserException

        /// <summary>
        /// GIF 파서 예외
        /// </summary>
        public class GIFParserException : Exception
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 위치 - Position

            /// <summary>
            /// 위치
            /// </summary>
            public int Position { get; private set; }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - GIFParserException(index, message)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="index"></param>
            /// <param name="message"></param>
            public GIFParserException(int index, string message) : base(message)
            {
                Position = index;
            }

            #endregion
            #region 생성자 - GIFParserException(index, message, innerException)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <param name="message">메시지</param>
            /// <param name="innerException">내부 예외</param>
            public GIFParserException(int index, string message, Exception innerException) : base(message, innerException)
            {
                Position = index;
            }

            #endregion
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 제로 파트 - ZeroPart

        /// <summary>
        /// 제로 파트
        /// </summary>
        protected class ZeroPart : GIFPart
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 길이 - Length

            /// <summary>
            /// 길이
            /// </summary>
            public override int Length
            {
                get
                {
                    return 0;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - ZeroPart()

            /// <summary>
            /// 생성자
            /// </summary>
            public ZeroPart()
            {
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 데이터 구하기 - GetData()

            /// <summary>
            /// 데이터 구하기
            /// </summary>
            /// <returns>데이터</returns>
            public override byte[] GetDataByteArray()
            {
                return new byte[0];
            }

            #endregion
            #region 복사 생성하기 - CreateCopy()

            /// <summary>
            /// 복사 생성하기
            /// </summary>
            /// <returns>GIF 파트</returns>
            public override GIFPart CreateCopy()
            {
                return new ZeroPart();
            }

            #endregion
        }

        #endregion
        #region 색상 테이블 파트 - ColorTablePart

        /// <summary>
        /// 색상 테이블 파트
        /// </summary>
        protected class ColorTablePart : GIFPartData, IList<Color>
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Class
            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 색상 테이블 열거자 - ColorTableEnumerator

            /// <summary>
            /// 색상 테이블 열거자
            /// </summary>
            private class ColorTableEnumerator : IEnumerator<Color>
            {
                //////////////////////////////////////////////////////////////////////////////////////////////////// Field
                ////////////////////////////////////////////////////////////////////////////////////////// Private

                #region Field

                /// <summary>
                /// 색상 테이블 파트
                /// </summary>
                private ColorTablePart table;

                /// <summary>
                /// 인덱스
                /// </summary>
                private int index;

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Property
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 현재 항목 - Current

                /// <summary>
                /// 현재 항목
                /// </summary>
                public Color Current
                {
                    get
                    {
                        return this.table[index];
                    }
                }

                #endregion
                #region 현재 항목 - IEnumerator.Current

                /// <summary>
                /// 현재 항목
                /// </summary>
                object IEnumerator.Current
                {
                    get
                    {
                        return Current;
                    }
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 생성자 - ColorTableEnumerator(table)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="table">색상 파트 테이블</param>
                public ColorTableEnumerator(ColorTablePart table)
                {
                    this.table = table;

                    this.index = -1;
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Method
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 리소스 해제하기 - Dispose()

                /// <summary>
                /// 리소스 해제하기
                /// </summary>
                public void Dispose()
                {
                }

                #endregion
                #region 다음 항목으로 이동하기 - MoveNext()

                /// <summary>
                /// 다음 항목으로 이동하기
                /// </summary>
                /// <returns>다음 항목 존재 여부</returns>
                public bool MoveNext()
                {
                    this.index++;

                    return this.index < this.table.Count;
                }

                #endregion
                #region 리셋하기 - Reset()

                /// <summary>
                /// 리셋하기
                /// </summary>
                public void Reset()
                {
                    this.index = -1;
                }

                #endregion
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Field
            ////////////////////////////////////////////////////////////////////////////////////////// Static
            //////////////////////////////////////////////////////////////////////////////// Private

            #region Field

            /// <summary>
            /// 유효 카운트 배열
            /// </summary>
            private static readonly int[] _validCountArray = new[] { 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080, 0x0100 };

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 카운트 - Count

            /// <summary>
            /// 카운트
            /// </summary>
            public int Count { get; private set; }

            #endregion
            #region 읽기 전용 여부 - IsReadOnly

            /// <summary>
            /// 읽기 전용 여부
            /// </summary>
            public bool IsReadOnly
            {
                get
                {
                    return true;
                }
            }

            #endregion
            #region 인덱서 - this[index]

            /// <summary>
            /// 인덱서
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <returns>색상</returns>
            public Color this[int index]
            {
                get
                {
                    if(index < 0 || index >= Count)
                    {
                        throw new IndexOutOfRangeException("Index is not within the color count range!");
                    }

                    int offset = 3 * index;

                    return Color.FromArgb(this.dataByteArray[offset + 0], this.dataByteArray[offset + 1], this.dataByteArray[offset + 2]);
                }
                set
                {
                    if(index < 0 || index >= Count)
                    {
                        throw new IndexOutOfRangeException("Index is not within the color count range!");
                    }

                    int offset = 3 * index;

                    this.dataByteArray[offset + 0] = value.R;
                    this.dataByteArray[offset + 1] = value.G;
                    this.dataByteArray[offset + 2] = value.B;
                }
            }
            
            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - ColorTablePart(dataByteArray, index, colorCount)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            /// <param name="colorCount">색상 수</param>
            public ColorTablePart(ref byte[] dataByteArray, ref int index, int colorCount) : base(ref dataByteArray, ref index, 3 * colorCount)
            {
                Count = colorCount;
            }

            #endregion
            #region 생성자 - ColorTablePart(colorList)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="colorList">색상 리스트</param>
            public ColorTablePart(List<Color> colorList)
            {
                if(colorList.Count > 0x0100)
                {
                    throw new Exception("Too many color for a GIF color table!");
                }

                while(!_validCountArray.Contains(colorList.Count))
                {
                    colorList.Add(Color.Black);
                }

                this.dataByteArray = new byte[3 * colorList.Count];

                int index = -1;

                foreach(Color color in colorList)
                {
                    this.dataByteArray[++index] = color.R;
                    this.dataByteArray[++index] = color.G;
                    this.dataByteArray[++index] = color.B;
                }

                Count = colorList.Count;
            }

            #endregion
            #region 생성자 - ColorTablePart(count, dataByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="count">카운트</param>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            private ColorTablePart(int count, byte[] dataByteArray)
            {
                Count = count;

                this.dataByteArray = dataByteArray;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 복사 생성하기 - CreateCopy()

            /// <summary>
            /// 복사 생성하기
            /// </summary>
            /// <returns>GIF 파트</returns>
            public override GIFPart CreateCopy()
            {
                byte[] bufferByteArray = new byte[this.dataByteArray.Length];

                Array.Copy(this.dataByteArray, bufferByteArray, this.dataByteArray.Length);

                return new ColorTablePart(this.Count, bufferByteArray);
            }

            #endregion
            #region 열거자 구하기 - GetEnumerator()

            /// <summary>
            /// 열거자 구하기
            /// </summary>
            /// <returns>열거자</returns>
            public IEnumerator<Color> GetEnumerator()
            {
                return new ColorTableEnumerator(this);
            }

            #endregion
            #region 열거자 구하기 - IEnumerable.GetEnumerator()

            /// <summary>
            /// 열거자 구하기
            /// </summary>
            /// <returns>열거자</returns>
            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }

            #endregion
            #region 추가하기 - Add(color)

            /// <summary>
            /// 추가하기
            /// </summary>
            /// <param name="color">색상</param>
            public void Add(Color color)
            {
                throw new Exception("This collection is read-only!");
            }

            #endregion
            #region 지우기 - Clear()

            /// <summary>
            /// 지우기
            /// </summary>
            public void Clear()
            {
                throw new Exception("This collection is read-only!");
            }

            #endregion
            #region 포함 여부 구하기 - Contains(color)

            /// <summary>
            /// 포함 여부 구하기
            /// </summary>
            /// <param name="color">색상</param>
            /// <returns>포함 여부</returns>
            public bool Contains(Color color)
            {
                bool result = false;

                for(int i = 0; i < Count; i++)
                {
                    if(this[i].Equals(color))
                    {
                        result = true;

                        break;
                    }
                }

                return result;
            }

            #endregion
            #region 복사하기 - CopyTo(colorArray, arrayIndex)

            /// <summary>
            /// 복사하기
            /// </summary>
            /// <param name="colorArray">색상 배열</param>
            /// <param name="arrayIndex">배열 인덱스</param>
            public void CopyTo(Color[] colorArray, int arrayIndex)
            {
                for(int i = 0; i < Count; i++)
                {
                    colorArray[arrayIndex + i] = this[i];
                }
            }

            #endregion
            #region 제거하기 - Remove(color)

            /// <summary>
            /// 제거하기
            /// </summary>
            /// <param name="color">색상</param>
            /// <returns>처리 결과</returns>
            public bool Remove(Color color)
            {
                throw new Exception("This collection is read-only!");
            }

            #endregion
            #region 인덱스 구하기 - IndexOf(color)

            /// <summary>
            /// 인덱스 구하기
            /// </summary>
            /// <param name="color">색상</param>
            /// <returns>인덱스</returns>
            public int IndexOf(Color color)
            {
                int result = -1;
 
                for(int i = 0; i < Count; i++)
                {
                    if(this[i].Equals(color))
                    {
                        result = i;

                        break;
                    }
                }

                return result;
            }

            #endregion
            #region 삽입하기 - Insert(index, color)

            /// <summary>
            /// 삽입하기
            /// </summary>
            /// <param name="index">인덱스</param>
            /// <param name="color">색상</param>
            public void Insert(int index, Color color)
            {
                throw new Exception("This collection is read-only!");
            }

            #endregion
            #region 제거하기 - RemoveAt(index)

            /// <summary>
            /// 제거하기
            /// </summary>
            /// <param name="index"></param>
            public void RemoveAt(int index)
            {
                throw new Exception("This collection is read-only!");
            }

            #endregion
        }

        #endregion
        #region GIF 데이터 섹션 - GIFDataSection

        /// <summary>
        /// GIF 데이터 섹션
        /// </summary>
        protected class GIFDataSection : GIFPart
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Class
            ////////////////////////////////////////////////////////////////////////////////////////// Protected

            #region 섹션 파트 - SectionPart

            /// <summary>
            /// 섹션 파트
            /// </summary>
            protected class SectionPart : GIFPartData
            {
                //////////////////////////////////////////////////////////////////////////////////////////////////// Property
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 데이터 바이트 배열 - DataByteArray

                /// <summary>
                /// 데이터 바이트 배열
                /// </summary>
                public byte[] DataByteArray
                {
                    get
                    {
                        return this.dataByteArray;
                    }
                }

                #endregion
                #region 데이터 길이 - DataLength

                /// <summary>
                /// 데이터 길이
                /// </summary>
                public byte DataLength
                {
                    get
                    {
                        return this.dataByteArray[0];
                    }
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 생성자 - SectionPart()

                /// <summary>
                /// 생성자
                /// </summary>
                public SectionPart()
                {
                    this.dataByteArray = new byte[] { 0x00 };
                }

                #endregion
                #region 생성자 - SectionPart(dataByteArray)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                public SectionPart(byte[] dataByteArray)
                {
                    if(dataByteArray.Length != dataByteArray[0] + 1)
                    {
                        throw new Exception("Length indicator does not equal the data length!");
                    }

                    this.dataByteArray = dataByteArray;
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Method
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 복사 생성하기 - CreateCopy()

                /// <summary>
                /// 복사 생성하기
                /// </summary>
                /// <returns>GIF 파트</returns>
                public override GIFPart CreateCopy()
                {
                    byte[] bufferByteArray = new byte[this.dataByteArray.Length];

                    Array.Copy(this.dataByteArray, bufferByteArray, this.dataByteArray.Length);

                    return new SectionPart(bufferByteArray);
                }

                #endregion
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 컨텐트 바이트 배열 - ContentByteArray

            /// <summary>
            /// 컨텐트 바이트 배열
            /// </summary>
            public byte[] ContentByteArray
            {
                get
                {
                    byte[] resultByteArray = null;

                    using(MemoryStream memoryStream = new MemoryStream())
                    {
                        this.partList.Select(p => p as SectionPart).ToList().ForEach(p => memoryStream.Write(p.DataByteArray, 1, p.DataLength));

                        resultByteArray = memoryStream.ToArray();
                    }

                    return resultByteArray;
                }
                set
                {
                    this.partList.Clear();

                    int index = 0;

                    byte[] bufferByteArray;

                    while(index < value.Length)
                    {
                        int count = Math.Min(255, value.Length - index);

                        bufferByteArray = new byte[count + 1];

                        bufferByteArray[0] = Convert.ToByte(count);

                        Array.Copy(value, index, bufferByteArray, 1, count);

                        this.partList.Add(new SectionPart(bufferByteArray));

                        index += count;
                    }

                    this.partList.Add(new SectionPart());
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - GIFDataSection(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public GIFDataSection(ref byte[] dataByteArray, ref int index)
            {
                int length;

                while((length = ((int)dataByteArray[index]) + 1) != 1)
                {
                    byte[] bufferByteArray = new byte[length];

                    Array.Copy(dataByteArray, index, bufferByteArray, 0, length);

                    this.partList.Add(new SectionPart(bufferByteArray));

                    index += length;
                }

                this.partList.Add(new SectionPart());

                index++;
            }

            #endregion
            #region 생성자 - GIFDataSection(contentByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="contentByteArray">컨텐트 바이트 배열</param>
            public GIFDataSection(byte[] contentByteArray)
            {
                ContentByteArray = contentByteArray;
            }

            #endregion
            #region 생성자 - GIFDataSection(partList)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="partList">GIF 파트 리스트</param>
            private GIFDataSection(List<GIFPart> partList)
            {
                this.partList = partList;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 복사 생성하기 - CreateCopy()

            /// <summary>
            /// 복사 생성하기
            /// </summary>
            /// <returns>GIF 파트</returns>
            public override GIFPart CreateCopy()
            {
                return new GIFDataSection(this.partList.Select(p => p.CreateCopy()).ToList());
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Interface
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region GIF 파트 컨텐트 인터페이스 - IGIFPartContent

        /// <summary>
        /// GIF 파트 컨텐트 인터페이스
        /// </summary>
        public interface IGIFPartContent
        {
            ////////////////////////////////////////////////////////////////////////////////////////// Property

            #region 컨텐트 - Content

            /// <summary>
            /// 컨텐트
            /// </summary>
            byte[] ContentByteArray { get; set; }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region Field

        /// <summary>
        /// 파트 리스트
        /// </summary>
        protected List<GIFPart> partList = new List<GIFPart>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 길이 - Length

        /// <summary>
        /// 길이
        /// </summary>
        public virtual int Length
        {
            get
            {
                return this.partList.Sum(p => p.Length);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 데이터 바이트 배열 구하기 - GetDataByteArray()

        /// <summary>
        /// 데이터 바이트 배열 구하기
        /// </summary>
        /// <returns>데이터 바이트 배열</returns>
        public virtual byte[] GetDataByteArray()
        {
            byte[] targetByteArray;

            using(MemoryStream memoryStream = new MemoryStream())
            {
                byte[] bufferByteArray;

                foreach(GIFPart part in this.partList)
                {
                    bufferByteArray = part.GetDataByteArray();

                    if(bufferByteArray != null && bufferByteArray.Length > 0)
                    {
                        memoryStream.Write(bufferByteArray, 0, bufferByteArray.Length);
                    }
                }

                targetByteArray = memoryStream.ToArray();
            }

            return targetByteArray;
        }

        #endregion
        #region 복사 생성하기 - CreateCopy()

        /// <summary>
        /// 복사 생성하기
        /// </summary>
        /// <returns>GIF 파트</returns>
        public abstract GIFPart CreateCopy();

        #endregion
    }
}