﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;

namespace TestProject
{
    /// <summary>
    /// GIF 파트 헤더
    /// </summary>
    public class GIFPartHeader : GIFPart
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 헤더 파트 - HeaderPart

        /// <summary>
        /// 헤더 파트
        /// </summary>
        private class HeaderPart : GIFPartData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Field
            ////////////////////////////////////////////////////////////////////////////////////////// Static
            //////////////////////////////////////////////////////////////////////////////// Private

            #region Field

            /// <summary>
            /// 적절한 버전 배열
            /// </summary>
            private static readonly string[] _validVersionArray = new[] { "89a", "87a" };

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 식별자 - Identifier

            /// <summary>
            /// 식별자
            /// </summary>
            public string Identifier
            {
                get
                {
                    return GetString(0, 3);
                }
                set
                {
                    SetString(0, 3, value);
                }
            }

            #endregion
            #region 버전 - Version

            /// <summary>
            /// 버전
            /// </summary>
            public string Version
            {
                get
                {
                    return GetString(3, 3);
                }
                set
                {
                    SetString(3, 3, value);
                }
            }

            #endregion
            #region 너비 - Width

            /// <summary>
            /// 너비
            /// </summary>
            public ushort Width
            {
                get
                {
                    return GetValue(6);
                }
                set
                {
                    SetValue(6, value);
                }
            }

            #endregion
            #region 높이 - Height

            /// <summary>
            /// 높이
            /// </summary>
            public ushort Height
            {
                get
                {
                    return GetValue(8);
                }
                set
                {
                    SetValue(8, value);
                }
            }

            #endregion
            #region 글로벌 색상 테이블 사용 여부 - UseGlobalColorTable

            /// <summary>
            /// 글로벌 색상 테이블 사용 여부
            /// </summary>
            public bool UseGlobalColorTable
            {
                get
                {
                    return (dataByteArray[10] & 0x80) != 0;
                }
                set
                {
                    dataByteArray[10] = Convert.ToByte((dataByteArray[10] & 0x7F) | (value ? 0x80 : 0x00));
                }
            }

            #endregion
            #region 색상 해상도 - ColorResolution

            /// <summary>
            /// 색상 해상도
            /// </summary>
            public byte ColorResolution
            {
                get
                {
                    return Convert.ToByte((dataByteArray[10] & 0x70) >> 4);
                }
                set
                {
                    dataByteArray[10] = Convert.ToByte((dataByteArray[10] & 0x8F) | ((value & 0x07) << 4));
                }
            }

            #endregion
            #region 테이블 정렬 여부 - TableSorted

            /// <summary>
            /// 테이블 정렬 여부
            /// </summary>
            public bool TableSorted
            {
                get
                {
                    return (dataByteArray[10] & 0x08) != 0;
                }
                set
                {
                    dataByteArray[10] = Convert.ToByte((dataByteArray[10] & 0xF7) | (value ? 0x04 : 0x00));
                }
            }

            #endregion
            #region 글로벌 색상 테이블 크기 - GlobalColorTableSize

            /// <summary>
            /// 글로벌 색상 테이블 크기
            /// </summary>
            public int GlobalColorTableSize
            {
                get
                {
                    return 2 << (dataByteArray[10] & 0x07);
                }
                set
                {
                    byte temporayValue = 0;

                    for(int i = 0; i < 7; i++)
                    {
                        if(value > (2 << i))
                        {
                            temporayValue++;
                        }
                    }

                    dataByteArray[10] = Convert.ToByte((dataByteArray[10] & 0xF8) | temporayValue);
                }
            }

            #endregion
            #region 배경 색상 - BackgroundColor

            /// <summary>
            /// 배경 색상
            /// </summary>
            public byte BackgroundColor
            {
                get
                {
                    return dataByteArray[11];
                }
                set
                {
                    dataByteArray[11] = value;
                }
            }

            #endregion
            #region 종횡비 - AspectRatio

            /// <summary>
            /// 종횡비
            /// </summary>
            public byte AspectRatio
            {
                get
                {
                    return dataByteArray[12];
                }
                set
                {
                    dataByteArray[12] = value;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - HeaderPart()

            /// <summary>
            /// 생성자
            /// </summary>
            public HeaderPart()
            {
                this.dataByteArray = new byte[13];

                Identifier = "GIF";

                Version = _validVersionArray[0];
            }

            #endregion
            #region 생성자 - HeaderPart(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public HeaderPart(ref byte[] dataByteArray, ref int index) : base(ref dataByteArray, ref index, 13)
            {
                if(!Identifier.Equals("GIF"))
                {
                    throw new GIFParserException(index - 13, "Header does not start with 'GIF'!");
                }

                if(!_validVersionArray.Any(v => v.Equals(Version)))
                {
                    throw new GIFParserException(index - 10, "File version not supported!");
                }
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 생성자 - HeaderPart(dataByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            private HeaderPart(byte[] dataByteArray)
            {
                base.dataByteArray = dataByteArray;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 복사 생성하기 - CreateCopy()

            /// <summary>
            /// 복사 생성하기
            /// </summary>
            /// <returns>GIF 파트</returns>
            public override GIFPart CreateCopy()
            {
                byte[] bufferByteArray = new byte[dataByteArray.Length];

                Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                return new HeaderPart(bufferByteArray);
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 헤더 - Header

        /// <summary>
        /// 헤더
        /// </summary>
        private HeaderPart Header
        {
            get
            {
                return this.partList[0] as HeaderPart;
            }
        }

        #endregion
        #region 식별자 - Identifier

        /// <summary>
        /// 식별자
        /// </summary>
        public string Identifier
        {
            get
            {
                return Header.Identifier;
            }
            set
            {
                Header.Identifier = value;
            }
        }

        #endregion
        #region 버전 - Version

        /// <summary>
        /// 버전
        /// </summary>
        public string Version
        {
            get
            {
                return Header.Version;
            }
            set
            {
                Header.Version = value;
            }
        }

        #endregion
        #region 너비 - Width

        /// <summary>
        /// 너비
        /// </summary>
        public ushort Width
        {
            get
            {
                return Header.Width;
            }
            set
            {
                Header.Width = value;
            }
        }

        #endregion
        #region 높이 - Height

        /// <summary>
        /// 높이
        /// </summary>
        public ushort Height
        {
            get
            {
                return Header.Height;
            }
            set
            {
                Header.Height = value;
            }
        }

        #endregion
        #region 글로벌 색상 테이블 사용 여부 - UseGlobalColorTable

        /// <summary>
        /// 글로벌 색상 테이블 사용 여부
        /// </summary>
        public bool UseGlobalColorTable
        {
            get
            {
                return Header.UseGlobalColorTable;
            }
            set
            {
                Header.UseGlobalColorTable = value;
            }
        }

        #endregion
        #region 색상 해상도 - ColorResolution

        /// <summary>
        /// 색상 해상도
        /// </summary>
        public byte ColorResolution
        {
            get
            {
                return Header.ColorResolution;
            }
            set
            {
                Header.ColorResolution = value;
            }
        }

        #endregion
        #region 테이블 정렬 여부 - TableSorted

        /// <summary>
        /// 테이블 정렬 여부
        /// </summary>
        public bool TableSorted
        {
            get
            {
                return Header.TableSorted;
            }
            set
            {
                Header.TableSorted = value;
            }
        }

        #endregion
        #region 글로벌 색상 테이블 크기 - GlobalColorTableSize

        /// <summary>
        /// 글로벌 색상 테이블 크기
        /// </summary>
        public int GlobalColorTableSize
        {
            get
            {
                return Header.GlobalColorTableSize;
            }
            set
            {
                Header.GlobalColorTableSize = value;
            }
        }

        #endregion
        #region 배경 색상 - BackgroundColor

        /// <summary>
        /// 배경 색상
        /// </summary>
        public byte BackgroundColor
        {
            get
            {
                return Header.BackgroundColor;
            }
            set
            {
                Header.BackgroundColor = value;
            }
        }

        #endregion
        #region 종횡비 - AspectRatio

        /// <summary>
        /// 종횡비
        /// </summary>
        public byte AspectRatio
        {
            get
            {
                return Header.AspectRatio;
            }
            set
            {
                Header.AspectRatio = value;
            }
        }

        #endregion
        #region 글로벌 색상 테이블 - GlobalColorTable

        /// <summary>
        /// 글로벌 색상 테이블
        /// </summary>
        public IList<Color> GlobalColorTable
        {
            get
            {
                return (partList[1] as ColorTablePart);
            }
            set
            {
                if(value == null)
                {
                    partList[1] = new ZeroPart();

                    UseGlobalColorTable = false;

                    GlobalColorTableSize = 0;
                }
                else
                {
                    ColorTablePart part = new ColorTablePart(value.ToList());

                    UseGlobalColorTable = true;

                    GlobalColorTableSize = part.Count;

                    partList[1] = part;
                }
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFPartHeader()

        /// <summary>
        /// 생성자
        /// </summary>
        public GIFPartHeader()
        {
            this.partList.Add(new HeaderPart());
            this.partList.Add(new ZeroPart());
        }

        #endregion
        #region 생성자 - GIFPartHeader(dataByteArray, index)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="dataByteArray">데이터 바이트 배열</param>
        /// <param name="index">인덱스</param>
        public GIFPartHeader(ref byte[] dataByteArray, ref int index)
        {
            this.partList.Add(new HeaderPart(ref dataByteArray, ref index));

            if(UseGlobalColorTable)
            {
                this.partList.Add(new ColorTablePart(ref dataByteArray, ref index, GlobalColorTableSize));
            }
            else
            {
                this.partList.Add(new ZeroPart());
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 생성자 - GIFPartHeader(partList)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="partList">GIF 파트 리스트</param>
        private GIFPartHeader(List<GIFPart> partList)
        {
            base.partList = partList;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 복사 생성하기 - CreateCopy()

        /// <summary>
        /// 복사 생성하기
        /// </summary>
        /// <returns>GIF 파트</returns>
        public override GIFPart CreateCopy()
        {
            return new GIFPartHeader(partList.Select(p => p.CreateCopy()).ToList());
        }

        #endregion
    }
}