﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;

namespace TestProject
{
    /// <summary>
    /// GIF 파트 LZW 이미지
    /// </summary>
    public class GIFPartLWZImage : GIFPart
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 헤더 파트 - HeaderPart

        /// <summary>
        /// 헤더 파트
        /// </summary>
        private class HeaderPart : GIFPartData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 식별자 - Identifier

            /// <summary>
            /// 식별자
            /// </summary>
            public byte Identifier
            {
                get
                {
                    return this.dataByteArray[0];
                }
                set
                {
                    this.dataByteArray[0] = value;
                }
            }

            #endregion
            #region 왼쪽 - Left

            /// <summary>
            /// 왼쪽
            /// </summary>
            public ushort Left
            {
                get
                {
                    return GetValue(1);
                }
                set
                {
                    SetValue(1, value);
                }
            }

            #endregion
            #region 위쪽 - Top

            /// <summary>
            /// 위쪽
            /// </summary>
            public ushort Top
            {
                get
                {
                    return GetValue(3);
                }
                set
                {
                    SetValue(3, value);
                }
            }

            #endregion
            #region 너비 - Width

            /// <summary>
            /// 너비
            /// </summary>
            public ushort Width
            {
                get
                {
                    return GetValue(5);
                }
                set
                {
                    SetValue(5, value);
                }
            }

            #endregion
            #region 높이 - Height

            /// <summary>
            /// 높이
            /// </summary>
            public ushort Height
            {
                get
                {
                    return GetValue(7);
                }
                set
                {
                    SetValue(7, value);
                }
            }

            #endregion
            #region 로컬 색상 테이블 사용 여부 - UseLocalColorTable

            /// <summary>
            /// 로컬 색상 테이블 사용 여부
            /// </summary>
            public bool UseLocalColorTable
            {
                get
                {
                    return (this.dataByteArray[9] & 0x80) != 0;
                }
                set
                {
                    this.dataByteArray[9] = Convert.ToByte((this.dataByteArray[9] & 0x7F) | (value ? 0x80 : 0x00));
                }
            }

            #endregion
            #region 인터레이스 여부 - Interlaced

            /// <summary>
            /// 인터레이스 여부
            /// </summary>
            public bool Interlaced
            {
                get
                {
                    return (this.dataByteArray[9] & 0x40) != 0;
                }
                set
                {
                    this.dataByteArray[9] = Convert.ToByte((this.dataByteArray[9] & 0xBF) | (value ? 0x40 : 0x00));
                }
            }

            #endregion
            #region 테이블 정렬 여부 - TableSorted

            /// <summary>
            /// 테이블 정렬 여부
            /// </summary>
            public bool TableSorted
            {
                get
                {
                    return (this.dataByteArray[9] & 0x20) != 0;
                }
                set
                {
                    this.dataByteArray[9] = Convert.ToByte((this.dataByteArray[9] & 0xDF) | (value ? 0x20 : 0x00));
                }
            }

            #endregion
            #region 로컬 색상 테이블 크기 - LocalColorTableSize

            /// <summary>
            /// 로컬 색상 테이블 크기
            /// </summary>
            public int LocalColorTableSize
            {
                get
                {
                    return 2 << (this.dataByteArray[9] & 0x07);
                }
                set
                {
                    byte temporaryValue = 0;

                    for(int i = 0; i < 7; i++)
                    {
                        if(value > (2 << i))
                        {
                            temporaryValue++;
                        }
                    }

                    this.dataByteArray[9] = Convert.ToByte((this.dataByteArray[9] & 0xF8) | temporaryValue);
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - HeaderPart()

            /// <summary>
            /// 생성자
            /// </summary>
            public HeaderPart()
            {
                this.dataByteArray = new byte[10];

                this.Identifier = 0x2C;
            }

            #endregion
            #region 생성자 - HeaderPart(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public HeaderPart(ref byte[] dataByteArray, ref int index) : base(ref dataByteArray, ref index, 10)
            {
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 생성자 - HeaderPart(dataByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            private HeaderPart(byte[] dataByteArray)
            {
                base.dataByteArray = dataByteArray;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 복사 생성하기 - CreateCopy()

            /// <summary>
            /// 복사 생성하기
            /// </summary>
            /// <returns>GIF 파트</returns>
            public override GIFPart CreateCopy()
            {
                byte[] bufferByteArray = new byte[dataByteArray.Length];

                Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                return new HeaderPart(bufferByteArray);
            }

            #endregion
        }

        #endregion
        #region 최소 코드 크기 파트 - GIFPartData

        /// <summary>
        /// 최소 코드 크기 파트
        /// </summary>
        private class MinimumCodeSizePart : GIFPartData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 최소 코드 크기 - MinimumCodeSize

            /// <summary>
            /// 최소 코드 크기
            /// </summary>
            public byte MinimumCodeSize
            {
                get
                {
                    return this.dataByteArray[0];
                }
                set
                {
                    this.dataByteArray[0] = value;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - MinimumCodeSizePart(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public MinimumCodeSizePart(ref byte[] dataByteArray, ref int index)
            {
                base.dataByteArray = new byte[] { dataByteArray[index] };

                index++;
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 생성자 - MinimumCodeSizePart(dataByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            private MinimumCodeSizePart(byte[] dataByteArray)
            {
                this.dataByteArray = dataByteArray;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 복사 생성하기 - CreateCopy()

            /// <summary>
            /// 복사 생성하기
            /// </summary>
            /// <returns>GIF 파트</returns>
            public override GIFPart CreateCopy()
            {
                byte[] bufferByteArray = new byte[dataByteArray.Length];

                Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                return new MinimumCodeSizePart(bufferByteArray);
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 레퍼 - Wrapper

        /// <summary>
        /// 레퍼
        /// </summary>
        public GIFWrapper Wrapper { get; set; }

        #endregion
        #region 헤더 - Header

        /// <summary>
        /// 헤더
        /// </summary>
        private HeaderPart Header
        {
            get
            {
                return this.partList[0] as HeaderPart;
            }
        }

        #endregion
        #region 식별자 - Identifier

        /// <summary>
        /// 식별자
        /// </summary>
        public byte Identifier
        {
            get
            {
                return Header.Identifier;
            }
            private set
            {
                Header.Identifier = value;
            }
        }

        #endregion
        #region 왼쪽 - Left

        /// <summary>
        /// 왼쪽
        /// </summary>
        public ushort Left
        {
            get
            {
                return Header.Left;
            }
            set
            {
                Header.Left = value;
            }
        }

        #endregion
        #region 위쪽 - Top

        /// <summary>
        /// 위쪽
        /// </summary>
        public ushort Top
        {
            get
            {
                return Header.Top;
            }
            set
            {
                Header.Top = value;
            }
        }

        #endregion
        #region 너비 - Width

        /// <summary>
        /// 너비
        /// </summary>
        public ushort Width
        {
            get
            {
                return Header.Width;
            }
            set
            {
                Header.Width = value;
            }
        }

        #endregion
        #region 높이 - Height

        /// <summary>
        /// 높이
        /// </summary>
        public ushort Height
        {
            get
            {
                return Header.Height;
            }
            set
            {
                Header.Height = value;
            }
        }

        #endregion
        #region 로컬 색상 테이블 사용 여부 - UseLocalColorTable

        /// <summary>
        /// 로컬 색상 테이블 사용 여부
        /// </summary>
        public bool UseLocalColorTable
        {
            get
            {
                return Header.UseLocalColorTable;
            }
            private set
            {
                Header.UseLocalColorTable = value;
            }
        }

        #endregion
        #region 인터레이스 여부 - Interlaced

        /// <summary>
        /// 인터레이스 여부
        /// </summary>
        public bool Interlaced
        {
            get
            {
                return Header.Interlaced;
            }
            set
            {
                Header.Interlaced = value;
            }
        }

        #endregion
        #region 테이블 정렬 여부 - TableSorted

        /// <summary>
        /// 테이블 정렬 여부
        /// </summary>
        public bool TableSorted
        {
            get
            {
                return Header.TableSorted;
            }
            set
            {
                Header.TableSorted = value;
            }
        }

        #endregion
        #region 로컬 색상 테이블 크기 - LocalColorTableSize

        /// <summary>
        /// 로컬 색상 테이블 크기
        /// </summary>
        public int LocalColorTableSize
        {
            get
            {
                return Header.LocalColorTableSize;
            }
            private set
            {
                Header.LocalColorTableSize = value;
            }
        }

        #endregion
        #region 로컬 색상 테이블 - LocalColorTable

        /// <summary>
        /// 로컬 색상 테이블
        /// </summary>
        public IList<Color> LocalColorTable
        {
            get
            {
                return (this.partList[1] as ColorTablePart);
            }
            set
            {
                if(value == null)
                {
                    this.partList[1] = new ZeroPart();

                    UseLocalColorTable = false;

                    LocalColorTableSize = 0;
                }
                else
                {
                    ColorTablePart part = new ColorTablePart(value.ToList());

                    UseLocalColorTable = true;

                    LocalColorTableSize = part.Count;

                    this.partList[1] = part;
                }
            }
        }

        #endregion
        #region 최소 코드 크기 파트 - MinCodeSizePart

        /// <summary>
        /// 최소 코드 크기 파트
        /// </summary>
        private MinimumCodeSizePart MinCodeSizePart
        {
            get
            {
                return this.partList[2] as MinimumCodeSizePart;
            }
        }

        #endregion
        #region 최소 코드 크기 - MinimumCodeSize

        /// <summary>
        /// 최소 코드 크기
        /// </summary>
        public byte MinimumCodeSize
        {
            get
            {
                return MinCodeSizePart.MinimumCodeSize;
            }
            set
            {
                MinCodeSizePart.MinimumCodeSize = value;
            }
        }

        #endregion
        #region 데이터 섹션 - DataSection

        /// <summary>
        /// 데이터 섹션
        /// </summary>
        private GIFDataSection DataSection
        {
            get
            {
                return this.partList[3] as GIFDataSection;
            }
        }

        #endregion
        #region 이미지 데이터 바이트 배열 - ImageDataByteArray

        /// <summary>
        /// 이미지 데이터 바이트 배열
        /// </summary>
        public byte[] ImageDataByteArray
        {
            get
            {
                return DataSection.ContentByteArray;
            }
            set
            {
                DataSection.ContentByteArray = value;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFPartLWZImage()

        /// <summary>
        /// 생성자
        /// </summary>
        public GIFPartLWZImage()
        {
            this.partList.Add(new HeaderPart());
        }

        #endregion
        #region 생성자 - GIFPartLWZImage(image)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="image">이미지</param>
        public GIFPartLWZImage(Image image)
        {
            GIFWrapper wrapper;

            using(MemoryStream memoryStream = new MemoryStream())
            {
                image.Save(memoryStream, ImageFormat.Gif);

                wrapper = new GIFWrapper(memoryStream.ToArray());
            }

            GIFPartHeader header = wrapper.Header;

            if(header == null)
            {
                throw new Exception("Could not load image as GIF!");
            }

            GIFPartLWZImage lwzImage = wrapper.FirstOrDefault(p => p is GIFPartLWZImage) as GIFPartLWZImage;

            if(lwzImage == null)
            {
                throw new Exception("Could not extract GIF image data from image!");
            }

            this.partList = lwzImage.partList;

            if(!UseLocalColorTable && header.UseGlobalColorTable)
            {
                LocalColorTable = header.GlobalColorTable;
            }
        }

        #endregion
        #region 생성자 - GIFPartLWZImage(wrapper, dataByteArray, index)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="wrapper">래퍼</param>
        /// <param name="dataByteArray">데이터 바이트 배열</param>
        /// <param name="index">인덱스</param>
        public GIFPartLWZImage(GIFWrapper wrapper, ref byte[] dataByteArray, ref int index)
        {
            this.Wrapper = wrapper;

            this.partList.Add(new HeaderPart(ref dataByteArray, ref index));

            if(UseLocalColorTable)
            {
                this.partList.Add(new ColorTablePart(ref dataByteArray, ref index, LocalColorTableSize));
            }
            else
            {
                this.partList.Add(new ZeroPart());
            }

            this.partList.Add(new MinimumCodeSizePart(ref dataByteArray, ref index));

            this.partList.Add(new GIFDataSection(ref dataByteArray, ref index));
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 생성자 - GIFPartLWZImage(partList)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="partList">GIF 파트 리스트</param>
        private GIFPartLWZImage(List<GIFPart> partList)
        {
            base.partList = partList;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region LWZ 이미지 파트 구하기 - GetLWZImagePart(image, colorResolution)

        /// <summary>
        /// LWZ 이미지 파트 구하기
        /// </summary>
        /// <param name="image">이미지</param>
        /// <param name="colorResolution">색상 해상도</param>
        /// <returns>GIF 파트 LWZ 이미지</returns>
        public static GIFPartLWZImage GetLWZImagePart(Image image, out byte colorResolution)
        {
            GIFWrapper wrapper;

            using(MemoryStream memoryStream = new MemoryStream())
            {
                image.Save(memoryStream, ImageFormat.Gif);

                wrapper = new GIFWrapper(memoryStream.ToArray());
            }

            GIFPartHeader header = wrapper.Header;

            if(header == null)
            {
                throw new Exception("Could not load image as GIF!");
            }

            colorResolution = header.ColorResolution;

            GIFPartLWZImage lwzImage = wrapper.FirstOrDefault(p => p is GIFPartLWZImage) as GIFPartLWZImage;

            if(lwzImage == null)
            {
                throw new Exception("Could not extract GIF image data from image!");
            }

            lwzImage.Wrapper = null;

            if(!lwzImage.UseLocalColorTable && header.UseGlobalColorTable)
            {
                lwzImage.LocalColorTable = header.GlobalColorTable;
            }

            return lwzImage;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 복사 생성하기 - CreateCopy()

        /// <summary>
        /// 복사 생성하기
        /// </summary>
        /// <returns>GIF 파트</returns>
        public override GIFPart CreateCopy()
        {
            return new GIFPartLWZImage(partList.Select(p => p.CreateCopy()).ToList());
        }

        #endregion
        #region 이미지 구하기 - ToImage()

        /// <summary>
        /// 이미지 구하기
        /// </summary>
        /// <returns>이미지</returns>
        public Image ToImage()
        {
            return Wrapper == null ? null : Wrapper.Extract(this).ToImage();
        }

        #endregion
    }
}