﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TestProject
{
    /// <summary>
    /// GIF 파트 메타 데이터
    /// </summary>
    public class GIFPartMetaData : GIFPart, GIFPart.IGIFPartContent
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 처분 방법 - DisposalMethod

        /// <summary>
        /// 처분 방법
        /// </summary>
        public enum DisposalMethod : byte
        {
            /// <summary>
            /// NotSpecified
            /// </summary>
            NotSpecified = 0x00,

            /// <summary>
            /// DoNotDispose
            /// </summary>
            DoNotDispose = 0x01,

            /// <summary>
            /// RestoreToBackgroundColor
            /// </summary>
            RestoreToBackgroundColor = 0x02,

            /// <summary>
            /// RestoreToPrevious
            /// </summary>
            RestoreToPrevious = 0x03,

            /// <summary>
            /// Reserved4
            /// </summary>
            Reserved4 = 0x04,

            /// <summary>
            /// Reserved5
            /// </summary>
            Reserved5 = 0x05,

            /// <summary>
            /// Reserved6
            /// </summary>
            Reserved6 = 0x06,

            /// <summary>
            /// Reserved7
            /// </summary>
            Reserved7 = 0x07
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Class
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 텍스트 그리기 - TextDraw

        /// <summary>
        /// 텍스트 그리기
        /// </summary>
        public class TextDraw : GIFPartMetaData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Class
            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 파트 - OptionPart

            /// <summary>
            /// 옵션 파트
            /// </summary>
            private class OptionPart : GIFPartData
            {
                //////////////////////////////////////////////////////////////////////////////////////////////////// Property
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 길이 - Length

                /// <summary>
                /// 길이
                /// </summary>
                public new byte Length
                {
                    get
                    {
                        return this.dataByteArray[0];
                    }
                    private set
                    {
                        this.dataByteArray[0] = value;
                    }
                }

                #endregion
                #region 왼쪽 - Left

                /// <summary>
                /// 왼쪽
                /// </summary>
                public ushort Left
                {
                    get
                    {
                        return GetValue(1);
                    }
                    set
                    {
                        SetValue(1, value);
                    }
                }

                #endregion
                #region 위쪽 - Top

                /// <summary>
                /// 위쪽
                /// </summary>
                public ushort Top
                {
                    get
                    {
                        return GetValue(3);
                    }
                    set
                    {
                        SetValue(3, value);
                    }
                }

                #endregion
                #region 너비 - Width

                /// <summary>
                /// 너비
                /// </summary>
                public ushort Width
                {
                    get
                    {
                        return GetValue(5);
                    }
                    set
                    {
                        SetValue(5, value);
                    }
                }

                #endregion
                #region 높이 - Height

                /// <summary>
                /// 높이
                /// </summary>
                public ushort Height
                {
                    get
                    {
                        return GetValue(7);
                    }
                    set
                    {
                        SetValue(7, value);
                    }
                }

                #endregion
                #region 문자 너비 - CharacterWidth

                /// <summary>
                /// 문자 너비
                /// </summary>
                public byte CharacterWidth
                {
                    get
                    {
                        return this.dataByteArray[9];
                    }
                    set
                    {
                        this.dataByteArray[9] = value;
                    }
                }

                #endregion
                #region 문자 높이 - CharacterHeight

                /// <summary>
                /// 문자 높이
                /// </summary>
                public byte CharacterHeight
                {
                    get
                    {
                        return this.dataByteArray[10];
                    }
                    set
                    {
                        this.dataByteArray[10] = value;
                    }
                }

                #endregion
                #region 텍스트 색상 - TextColor

                /// <summary>
                /// 텍스트 색상
                /// </summary>
                public byte TextColor
                {
                    get
                    {
                        return this.dataByteArray[11];
                    }
                    set
                    {
                        this.dataByteArray[11] = value;
                    }
                }

                #endregion
                #region 배경 색상 - BackgroundColor

                /// <summary>
                /// 배경 색상
                /// </summary>
                public byte BackgroundColor
                {
                    get
                    {
                        return this.dataByteArray[12];
                    }
                    set
                    {
                        this.dataByteArray[12] = value;
                    }
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 생성자 - OptionPart()

                /// <summary>
                /// 생성자
                /// </summary>
                public OptionPart()
                {
                    this.dataByteArray = new byte[13];

                    this.dataByteArray[0] = 12;
                }

                #endregion
                #region 생성자 - OptionPart(dataByteArray, index)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                /// <param name="index">인덱스</param>
                public OptionPart(ref byte[] dataByteArray, ref int index) : base(ref dataByteArray, ref index, dataByteArray[index] + 1)
                {
                }

                #endregion

                ////////////////////////////////////////////////////////////////////////////////////////// Private

                #region 생성자 - OptionPart(dataByteArray)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                private OptionPart(byte[] dataByteArray)
                {
                    this.dataByteArray = dataByteArray;
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Method
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 복사 생성하기 - CreateCopy()

                /// <summary>
                /// 복사 생성하기
                /// </summary>
                /// <returns>GIF 파트</returns>
                public override GIFPart CreateCopy()
                {
                    byte[] bufferByteArray = new byte[dataByteArray.Length];

                    Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                    return new OptionPart(bufferByteArray);
                }

                #endregion
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 왼쪽 - Left

            /// <summary>
            /// 왼쪽
            /// </summary>
            public ushort Left
            {
                get
                {
                    return Option.Left;
                }
                set
                {
                    Option.Left = value;
                }
            }

            #endregion
            #region 위쪽 - Top

            /// <summary>
            /// 위쪽
            /// </summary>
            public ushort Top
            {
                get
                {
                    return Option.Top;
                }
                set
                {
                    Option.Top = value;
                }
            }

            #endregion
            #region 너비 - Width

            /// <summary>
            /// 너비
            /// </summary>
            public ushort Width
            {
                get
                {
                    return Option.Width;
                }
                set
                {
                    Option.Width = value;
                }
            }

            #endregion
            #region 높이 - Height

            /// <summary>
            /// 높이
            /// </summary>
            public ushort Height
            {
                get
                {
                    return Option.Height;
                }
                set
                {
                    Option.Height = value;
                }
            }

            #endregion
            #region 문자 너비 - CharacterWidth

            /// <summary>
            /// 문자 너비
            /// </summary>
            public byte CharacterWidth
            {
                get
                {
                    return Option.CharacterWidth;
                }
                set
                {
                    Option.CharacterWidth = value;
                }
            }

            #endregion
            #region 문자 높이 - CharacterHeight

            /// <summary>
            /// 문자 높이
            /// </summary>
            public byte CharacterHeight
            {
                get
                {
                    return Option.CharacterHeight;
                }
                set
                {
                    Option.CharacterHeight = value;
                }
            }

            #endregion
            #region 텍스트 색상 - TextColor

            /// <summary>
            /// 텍스트 색상
            /// </summary>
            public byte TextColor
            {
                get
                {
                    return Option.TextColor;
                }
                set
                {
                    Option.TextColor = value;
                }
            }

            #endregion
            #region 배경 색상 - BackgroundColor

            /// <summary>
            /// 배경 색상
            /// </summary>
            public byte BackgroundColor
            {
                get
                {
                    return Option.BackgroundColor;
                }
                set
                {
                    Option.BackgroundColor = value;
                }
            }

            #endregion
            #region 텍스트 - Text

            /// <summary>
            /// 텍스트
            /// </summary>
            public string Text
            {
                get
                {
                    return Encoding.ASCII.GetString(Data.ContentByteArray);
                }
                set
                {
                    Data.ContentByteArray = Encoding.ASCII.GetBytes(value);
                }
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 - Option

            /// <summary>
            /// 옵션
            /// </summary>
            private OptionPart Option
            {
                get
                {
                    return this.partList[1] as OptionPart;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - TextDraw(text)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="text">텍스트</param>
            public TextDraw(string text)
            {
                this.partList.Add(new HeaderPart(0x01));

                this.partList.Add(new OptionPart());

                this.partList.Add(new GIFDataSection(Encoding.ASCII.GetBytes(text)));
            }

            #endregion
            #region 생성자 - TextDraw(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public TextDraw(ref byte[] dataByteArray, ref int index) : base()
            {
                HeaderPart header = new HeaderPart(ref dataByteArray, ref index);

                if(header.MetaType != 0x01)
                {
                    throw new GIFParserException(index - 2, "MetaType does not match the TextMeta-Type 0x01!");
                }

                this.partList.Add(header);

                if(dataByteArray[index] < 12)
                {
                    throw new GIFParserException(index, "Text does not hat the required option length!");
                }

                this.partList.Add(new OptionPart(ref dataByteArray, ref index));

                this.partList.Add(new GIFDataSection(ref dataByteArray, ref index));
            }

            #endregion
        }

        #endregion
        #region 그래픽스 컨트롤 - GraphicsControl

        /// <summary>
        /// 그래픽스 컨트롤
        /// </summary>
        public class GraphicsControl : GIFPartMetaData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Class
            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 파트 - OptionPart

            /// <summary>
            /// 옵션 파트
            /// </summary>
            private class OptionPart : GIFPartData
            {
                //////////////////////////////////////////////////////////////////////////////////////////////////// Property
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 길이 - Length

                /// <summary>
                /// 길이
                /// </summary>
                public new byte Length
                {
                    get
                    {
                        return this.dataByteArray[0];
                    }
                    private set
                    {
                        this.dataByteArray[0] = value;
                    }
                }

                #endregion
                #region 처분 방법 - DisposalMethod

                /// <summary>
                /// 처분 방법
                /// </summary>
                public byte DisposalMethod
                {
                    get
                    {
                        return Convert.ToByte((this.dataByteArray[1] & 0x1C) >> 2);
                    }
                    set
                    {
                        this.dataByteArray[1] = Convert.ToByte((this.dataByteArray[1] & 0xE3) | ((value & 0x07) << 2));
                    }
                }

                #endregion
                #region 사용자 입력 사용 여부 - UseUserInput

                /// <summary>
                /// 사용자 입력 사용 여부
                /// </summary>
                public bool UseUserInput
                {
                    get
                    {
                        return (this.dataByteArray[1] & 0x02) != 0;
                    }
                    set
                    {
                        this.dataByteArray[1] = Convert.ToByte((this.dataByteArray[1] & 0xFD) | (value ? 0x02 : 0x00));
                    }
                }

                #endregion
                #region 투명도 사용 여부 - UseTransparency

                /// <summary>
                /// 투명도 사용 여부
                /// </summary>
                public bool UseTransparency
                {
                    get
                    {
                        return (this.dataByteArray[1] & 0x01) != 0;
                    }
                    set
                    {
                        this.dataByteArray[1] = Convert.ToByte((this.dataByteArray[1] & 0xFE) | (value ? 0x01 : 0x00));
                    }
                }

                #endregion
                #region 지연 시간 - DelayTime

                /// <summary>
                /// 지연 시간
                /// </summary>
                public ushort DelayTime
                {
                    get
                    {
                        return GetValue(2);
                    }
                    set
                    {
                        SetValue(2, value);
                    }
                }

                #endregion
                #region 투명 색상 - TransparentColor

                /// <summary>
                /// 투명 색상
                /// </summary>
                public byte TransparentColor
                {
                    get
                    {
                        return this.dataByteArray[4];
                    }
                    set
                    {
                        this.dataByteArray[4] = value;
                    }
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 생성자 - OptionPart()

                /// <summary>
                /// 생성자
                /// </summary>
                public OptionPart()
                {
                    this.dataByteArray = new byte[6];

                    this.dataByteArray[0] = 4;
                    this.dataByteArray[5] = 0;
                }

                #endregion
                #region 생성자 - OptionPart(dataByteArray, index)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                /// <param name="index">인덱스</param>
                public OptionPart(ref byte[] dataByteArray, ref int index) : base(ref dataByteArray, ref index, dataByteArray[index] + 2)
                {
                    if(base.dataByteArray[base.dataByteArray.Length - 1] != 0x00)
                    {
                        throw new GIFParserException(index - 1, "The Graphics Control meta must not contain more than one data section!");
                    }
                }

                #endregion

                ////////////////////////////////////////////////////////////////////////////////////////// Private

                #region 생성자 - OptionPart(dataByteArray)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                private OptionPart(byte[] dataByteArray)
                {
                    this.dataByteArray = dataByteArray;
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Method
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 복사 생성하기 - CreateCopy()

                /// <summary>
                /// 복사 생성하기
                /// </summary>
                /// <returns>GIF 파트</returns>
                public override GIFPart CreateCopy()
                {
                    byte[] bufferByteArray = new byte[dataByteArray.Length];

                    Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                    return new OptionPart(bufferByteArray);
                }

                #endregion
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 처분 방법 - DisposalMethod

            /// <summary>
            /// 처분 방법
            /// </summary>
            public new DisposalMethod DisposalMethod
            {
                get
                {
                    return (DisposalMethod)Option.DisposalMethod;
                }
                set
                {
                    Option.DisposalMethod = (byte)value;
                }
            }

            #endregion
            #region 사용자 입력 사용 여부 - UseUserInput

            /// <summary>
            /// 사용자 입력 사용 여부
            /// </summary>
            public bool UseUserInput
            {
                get
                {
                    return Option.UseUserInput;
                }
                set
                {
                    Option.UseUserInput = value;
                }
            }

            #endregion
            #region 투명도 사용 여부 - UseTransparency

            /// <summary>
            /// 투명도 사용 여부
            /// </summary>
            public bool UseTransparency
            {
                get
                {
                    return Option.UseTransparency;
                }
                set
                {
                    Option.UseTransparency = value;
                }
            }

            #endregion
            #region 지연 시간 - DelayTime

            /// <summary>
            /// 지연 시간
            /// </summary>
            public ushort DelayTime
            {
                get
                {
                    return Option.DelayTime;
                }
                set
                {
                    Option.DelayTime = value;
                }
            }

            #endregion
            #region 투명 색상 - TransparentColor

            /// <summary>
            /// 투명 색상
            /// </summary>
            public byte TransparentColor
            {
                get
                {
                    return Option.TransparentColor;
                }
                set
                {
                    Option.TransparentColor = value;
                }
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 - Option

            /// <summary>
            /// 옵션
            /// </summary>
            private OptionPart Option
            {
                get
                {
                    return this.partList[1] as OptionPart;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - GraphicsControl()

            /// <summary>
            /// 생성자
            /// </summary>
            public GraphicsControl()
            {
                this.partList.Add(new HeaderPart(0xF9));

                this.partList.Add(new OptionPart());
            }

            #endregion
            #region 생성자 - GraphicsControl(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public GraphicsControl(ref byte[] dataByteArray, ref int index)
            {
                HeaderPart header = new HeaderPart(ref dataByteArray, ref index);

                this.partList.Add(header);

                if(header.MetaType != 0xF9)
                {
                    throw new GIFParserException(index - 2, "The Graphics Control meta type does not match 0xF9!");
                }

                if(dataByteArray[index] < 4)
                {
                    throw new GIFParserException(index, "The Graphics Control meta data is too short!");
                }

                this.partList.Add(new OptionPart(ref dataByteArray, ref index));
            }

            #endregion
        }

        #endregion
        #region 주석 - Comment

        /// <summary>
        /// 주석
        /// </summary>
        public class Comment : GIFPartMetaData, IGIFPartContent
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - Comment(contentByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="contentByteArray">컨텐트 바이트 배열</param>
            public Comment(byte[] contentByteArray)
            {
                this.partList.Add(new HeaderPart(0xFE));

                this.partList.Add(new GIFDataSection(contentByteArray));
            }

            #endregion
            #region 생성자 - Comment(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public Comment(ref byte[] dataByteArray, ref int index) : base(ref dataByteArray, ref index)
            {
                if(MetaType != 0xFE)
                {
                    throw new GIFParserException(index - 1, "The comment meta type must be 0xFE!");
                }
            }

            #endregion
        }

        #endregion
        #region 어플리케이션 데이터 - ApplicationData

        /// <summary>
        /// 어플리케이션 데이터
        /// </summary>
        public class ApplicationData : GIFPartMetaData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Class
            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 파트 - OptionPart

            /// <summary>
            /// 옵션 파트
            /// </summary>
            private class OptionPart : GIFPartData
            {
                //////////////////////////////////////////////////////////////////////////////////////////////////// Property
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 앱 식별자 - AppIdentifier

                /// <summary>
                /// 앱 식별자
                /// </summary>
                public string AppIdentifier
                {
                    get
                    {
                        return Encoding.ASCII.GetString(this.dataByteArray, 1, this.dataByteArray.Length - 1);
                    }
                    set
                    {
                        byte[] bufferByteArray = Encoding.ASCII.GetBytes(value);

                        if(bufferByteArray.Length > 255)
                        {
                            throw new Exception("The AppIdent must not be longer than 255 characters!");
                        }

                        this.dataByteArray = new byte[bufferByteArray.Length + 1];

                        this.dataByteArray[0] = Convert.ToByte(bufferByteArray.Length);

                        Array.Copy(bufferByteArray, 0, this.dataByteArray, 1, bufferByteArray.Length);
                    }
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 생성자 - OptionPart()

                /// <summary>
                /// 생성자
                /// </summary>
                public OptionPart()
                {
                    this.dataByteArray = new byte[12];

                    this.dataByteArray[0] = 11;
                }

                #endregion
                #region 생성자 - OptionPart(dataByteArray, index)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                /// <param name="index">인덱스</param>
                public OptionPart(ref byte[] dataByteArray, ref int index) : base(ref dataByteArray, ref index, dataByteArray[index] + 1)
                {
                }

                #endregion

                ////////////////////////////////////////////////////////////////////////////////////////// Private

                #region 생성자 - OptionPart(dataByteArray)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                private OptionPart(byte[] dataByteArray)
                {
                    this.dataByteArray = dataByteArray;
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Method
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 복사 생성하기 - CreateCopy()

                /// <summary>
                /// 복사 생성하기
                /// </summary>
                /// <returns>GIF 파트</returns>
                public override GIFPart CreateCopy()
                {
                    byte[] bufferByteArray = new byte[dataByteArray.Length];

                    Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                    return new OptionPart(bufferByteArray);
                }

                #endregion
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 앱 식별자 - AppIdentifier

            /// <summary>
            /// 앱 식별자
            /// </summary>
            public string AppIdentifier
            {
                get
                {
                    return Option.AppIdentifier;
                }
                set
                {
                    Option.AppIdentifier = value;
                }
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 - Option

            /// <summary>
            /// 옵션
            /// </summary>
            private OptionPart Option
            {
                get
                {
                    return this.partList[1] as OptionPart;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - ApplicationData(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public ApplicationData(ref byte[] dataByteArray, ref int index) : base()
            {
                HeaderPart header = new HeaderPart(ref dataByteArray, ref index);

                this.partList.Add(header);

                if(MetaType != 0xFF)
                {
                    throw new GIFParserException(index - 1, "The Application meta type does not match 0xFF!");
                }

                this.partList.Add(new OptionPart(ref dataByteArray, ref index));

                this.partList.Add(new GIFDataSection(ref dataByteArray, ref index));
            }

            #endregion
        }

        #endregion
        #region 루프 - Loop

        /// <summary>
        /// 루프
        /// </summary>
        public class Loop : GIFPartMetaData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Class
            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 파트 - OptionPart

            /// <summary>
            /// 옵션 파트
            /// </summary>
            private class OptionPart : GIFPartData
            {
                //////////////////////////////////////////////////////////////////////////////////////////////////// Property
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 길이 - Length

                /// <summary>
                /// 길이
                /// </summary>
                public override int Length
                {
                    get
                    {
                        return 11;
                    }
                }

                #endregion
                #region 앱 식별자 - AppIdentifier

                /// <summary>
                /// 앱 식별자
                /// </summary>
                public string AppIdentifier
                {
                    get
                    {
                        return "NETSCAPE2.0";
                    }
                    set
                    {
                    }
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 생성자 - OptionPart()

                /// <summary>
                /// 생성자
                /// </summary>
                public OptionPart()
                {
                    this.dataByteArray = new byte[12];

                    this.dataByteArray[0] = 11;

                    byte[] bufferByteArray = Encoding.ASCII.GetBytes(AppIdentifier);

                    Array.Copy(bufferByteArray, 0, dataByteArray, 1, 11);
                }

                #endregion

                ////////////////////////////////////////////////////////////////////////////////////////// Private

                #region 생성자 - OptionPart(dataByteArray)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                private OptionPart(byte[] dataByteArray)
                {
                    this.dataByteArray = dataByteArray;
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Method
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 복사 생성하기 - CreateCopy()

                /// <summary>
                /// 복사 생성하기
                /// </summary>
                /// <returns>GIF 파트</returns>
                public override GIFPart CreateCopy()
                {
                    byte[] bufferByteArray = new byte[dataByteArray.Length];

                    Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                    return new OptionPart(bufferByteArray);
                }

                #endregion
            }

            #endregion
            #region 매개 변수 파트 - ParameterPart

            /// <summary>
            /// 매개 변수 파트
            /// </summary>
            private class ParameterPart : GIFPartData
            {
                //////////////////////////////////////////////////////////////////////////////////////////////////// Property
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 반복 수 - IterationCount

                /// <summary>
                /// 반복 수
                /// </summary>
                public ushort IterationCount
                {
                    get
                    {
                        return GetValue(2);
                    }
                    set
                    {
                        SetValue(2, value);
                    }
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 생성자 - ParameterPart()

                /// <summary>
                /// 생성자
                /// </summary>
                public ParameterPart()
                {
                    this.dataByteArray = new byte[5];

                    this.dataByteArray[0] = 3;
                    this.dataByteArray[1] = 1;
                    this.dataByteArray[2] = 0;
                    this.dataByteArray[3] = 0;
                    this.dataByteArray[4] = 0;
                }

                #endregion
                #region 생성자 - ParameterPart(originalDataByteArray)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="originalDataByteArray">원본 데이터 바이트 배열</param>
                public ParameterPart(byte[] originalDataByteArray)
                {
                    if(originalDataByteArray.Length != 3)
                    {
                        throw new GIFParserException(-1, "The Loop Application parameter data must be 3 bytes!");
                    }

                    this.dataByteArray = new byte[5];

                    this.dataByteArray[0] = 3;
                    this.dataByteArray[1] = 1;
                    this.dataByteArray[2] = originalDataByteArray[1];
                    this.dataByteArray[3] = originalDataByteArray[2];
                    this.dataByteArray[4] = 0;
                }

                #endregion

                ////////////////////////////////////////////////////////////////////////////////////////// Private

                #region 생성자 - ParameterPart(dataByteArray, noop)

                /// <summary>
                /// 생성자
                /// </summary>
                /// <param name="dataByteArray">데이터 바이트 배열</param>
                /// <param name="noop">NOOP</param>
                private ParameterPart(byte[] dataByteArray, bool noop)
                {
                    base.dataByteArray = dataByteArray;
                }

                #endregion

                //////////////////////////////////////////////////////////////////////////////////////////////////// Method
                ////////////////////////////////////////////////////////////////////////////////////////// Public

                #region 복사 생성하기 - CreateCopy()

                /// <summary>
                /// 복사 생성하기
                /// </summary>
                /// <returns>GIF 파트</returns>
                public override GIFPart CreateCopy()
                {
                    byte[] bufferByteArray = new byte[dataByteArray.Length];

                    Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                    return new ParameterPart(bufferByteArray, true);
                }

                #endregion
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 앱 식별자 - AppIdentifier

            /// <summary>
            /// 앱 식별자
            /// </summary>
            public string AppIdentifier
            {
                get
                {
                    return Option.AppIdentifier;
                }
            }

            #endregion
            #region 반복 수 - IterationCount

            /// <summary>
            /// 반복 수
            /// </summary>
            public ushort IterationCount
            {
                get
                {
                    return Parameter.IterationCount;
                }
                set
                {
                    Parameter.IterationCount = value;
                }
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 옵션 - Option

            /// <summary>
            /// 옵션
            /// </summary>
            private OptionPart Option
            {
                get
                {
                    return this.partList[1] as OptionPart;
                }
            }

            #endregion
            #region 매개 변수 - Parameter

            /// <summary>
            /// 매개 변수
            /// </summary>
            private ParameterPart Parameter
            {
                get
                {
                    return this.partList[2] as ParameterPart;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - Loop()

            /// <summary>
            /// 생성자
            /// </summary>
            public Loop()
            {
                this.partList.Add(new HeaderPart(0xFF));

                this.partList.Add(new OptionPart());

                this.partList.Add(new ParameterPart());
            }

            #endregion
            #region 생성자 - Loop(data)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="data">어플리케이션 데이터</param>
            public Loop(ApplicationData data)
            {
                this.partList.Add(new HeaderPart(0xFF));

                this.partList.Add(new OptionPart());

                this.partList.Add(new ParameterPart(data.ContentByteArray));
            }

            #endregion
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 헤더 파트 - HeaderPart

        /// <summary>
        /// 헤더 파트
        /// </summary>
        protected class HeaderPart : GIFPartData
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////// Property
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 식별자 - Identifier

            /// <summary>
            /// 식별자
            /// </summary>
            public byte Identifier
            {
                get
                {
                    return this.dataByteArray[0];
                }
                set
                {
                    this.dataByteArray[0] = value;
                }
            }

            #endregion
            #region 메타 타입 - MetaType

            /// <summary>
            /// 메타 타입
            /// </summary>
            public byte MetaType
            {
                get
                {
                    return this.dataByteArray[1];
                }
                set
                {
                    this.dataByteArray[1] = value;
                }
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 생성자 - HeaderPart(metaType)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="metaType">메타 타입</param>
            public HeaderPart(byte metaType)
            {
                this.dataByteArray = new byte[] { 0x21, metaType };
            }

            #endregion
            #region 생성자 - HeaderPart(dataByteArray, index)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            /// <param name="index">인덱스</param>
            public HeaderPart(ref byte[] dataByteArray, ref int index) : base(ref dataByteArray, ref index, 2)
            {
            }

            #endregion

            ////////////////////////////////////////////////////////////////////////////////////////// Private

            #region 생성자 - HeaderPart(dataByteArray)

            /// <summary>
            /// 생성자
            /// </summary>
            /// <param name="dataByteArray">데이터 바이트 배열</param>
            private HeaderPart(byte[] dataByteArray)
            {
                this.dataByteArray = dataByteArray;
            }

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////// Method
            ////////////////////////////////////////////////////////////////////////////////////////// Public

            #region 복사 생성하기 - CreateCopy()

            /// <summary>
            /// 복사 생성하기
            /// </summary>
            /// <returns>GIF 파트</returns>
            public override GIFPart CreateCopy()
            {
                byte[] bufferByteArray = new byte[dataByteArray.Length];

                Array.Copy(dataByteArray, bufferByteArray, dataByteArray.Length);

                return new HeaderPart(bufferByteArray);
            }

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 헤더 - Header

        /// <summary>
        /// 헤더
        /// </summary>
        private HeaderPart Header
        {
            get
            {
                return this.partList[0] as HeaderPart;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 데이터 - Data

        /// <summary>
        /// 데이터
        /// </summary>
        protected GIFDataSection Data
        {
            get
            {
                return this.partList[this.partList.Count - 1] as GIFDataSection;
            }
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 식별자 - Identifier

        /// <summary>
        /// 식별자
        /// </summary>
        public byte Identifier
        {
            get
            {
                return Header.Identifier;
            }
            set
            {
                Header.Identifier = value;
            }
        }

        #endregion
        #region 메타 타입 - MetaType

        /// <summary>
        /// 메타 타입
        /// </summary>
        public byte MetaType
        {
            get
            {
                return Header.MetaType;
            }
            set
            {
                Header.MetaType = value;
            }
        }

        #endregion
        #region 컨텐트 바이트 배열 - ContentByteArray

        /// <summary>
        /// 컨텐트 바이트 배열
        /// </summary>
        public byte[] ContentByteArray
        {
            get
            {
                return Data == null ? null : Data.ContentByteArray;
            }
            set
            {
                if(Data != null)
                {
                    Data.ContentByteArray = value;
                }
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFPartMetaData(metaType, dataByteArray)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="metaType">메타 타입</param>
        /// <param name="dataByteArray">데이터 바이트 배열</param>
        public GIFPartMetaData(byte metaType, byte[] dataByteArray)
        {
            this.partList.Add(new HeaderPart(metaType));

            this.partList.Add(new GIFDataSection(dataByteArray));
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 생성자 - GIFPartMetaData()

        /// <summary>
        /// 생성자
        /// </summary>
        protected GIFPartMetaData()
        {
        }

        #endregion
        #region 생성자 - GIFPartMetaData(dataByteArray, index)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="dataByteArray">데이터 바이트 배열</param>
        /// <param name="index">인덱스</param>
        protected GIFPartMetaData(ref byte[] dataByteArray, ref int index)
        {
            this.partList.Add(new HeaderPart(ref dataByteArray, ref index));

            this.partList.Add(new GIFDataSection(ref dataByteArray, ref index));
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 생성자 - GIFPartMetaData(partList)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="partList">GIF 파트 리스트</param>
        private GIFPartMetaData(List<GIFPart> partList)
        {
            base.partList = partList;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 파싱하기 - Parse(dataByteArray, index)

        /// <summary>
        /// 파싱하기
        /// </summary>
        /// <param name="dataByteArray">데이터 바이트 배열</param>
        /// <param name="index">인덱스</param>
        /// <returns>GIF 파트 메타 데이터</returns>
        public static GIFPartMetaData Parse(ref byte[] dataByteArray, ref int index)
        {
            if(dataByteArray[index] != 0x21)
            {
                throw new GIFParserException(index, string.Format("Not the start of a meta data part: {0:X2}", dataByteArray[index]));
            }

            GIFPartMetaData metaData = null;

            switch(dataByteArray[index + 1])
            {
                case 0x01 :
                
                    metaData = new TextDraw(ref dataByteArray, ref index);
                    
                    break;

                case 0xF9 :
                
                    metaData = new GraphicsControl(ref dataByteArray, ref index);
                    
                    break;

                case 0xFE :
                
                    metaData = new Comment(ref dataByteArray, ref index);
                    
                        break;

                case 0xFF :
                
                    ApplicationData applicationData = new ApplicationData(ref dataByteArray, ref index);

                    if(applicationData.AppIdentifier.Equals("NETSCAPE2.0"))
                    {
                        metaData = new Loop(applicationData);
                    }
                    else
                    {
                        metaData = applicationData;
                    }

                    break;

                default :

                    metaData = new GIFPartMetaData(ref dataByteArray, ref index);

                    break;
            }

            return metaData;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 복사 생성하기 - CreateCopy()

        /// <summary>
        /// 복사 생성하기
        /// </summary>
        /// <returns>GIF 파트</returns>
        public override GIFPart CreateCopy()
        {
            return new GIFPartMetaData(partList.Select(p => p.CreateCopy()).ToList());
        }

        #endregion
    }
}