﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// GIF 색상 테이블 변환 폼
    /// </summary>
    public partial class GIFColorTableTransformForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 변경 여부
        /// </summary>
        private bool isChanged = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 값 - Value

        /// <summary>
        /// 값
        /// </summary>
        public IList<Color> Value
        {
            get
            {
                return this.targetColorTableControl.Value;
            }
            set
            {
                this.sourceColorTableControl.Value = value;

                this.targetColorTableControl.Value = value;

                SetTargetColorTable();
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFColorTableTransformForm(colorTable)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="colorTable">색상 테이블</param>
        public GIFColorTableTransformForm(IList<Color> colorTable)
        {
            InitializeComponent();

            Value = colorTable;

            #region 이벤트를 설정한다.

            this.brightnessTrackBar.ValueChanged      += brightnessTrackBar_ValueChanged;
            this.brightnessNumericUpDown.ValueChanged += brightnessNumericUpDown_ValueChanged;
            this.contrastNumericUpDown.ValueChanged   += contrastNumericUpDown_ValueChanged;
            this.contrastTrackBar.ValueChanged        += contrastTrackBar_ValueChanged;
            this.colorTrackBar.ValueChanged           += colorTrackBar_ValueChanged;
            this.colorNumericUpDown.ValueChanged      += colorNumericUpDown_ValueChanged;
            this.applyButton.Click                    += applyButton_Click;
            this.cancelButton.Click                   += cancelButton_Click;

            #endregion
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 명도 트랙바 값 변경시 처리하기 - brightnessTrackBar_ValueChanged(sender, e)

        /// <summary>
        /// 명도 트랙바 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void brightnessTrackBar_ValueChanged(object sender, EventArgs e)
        {
            if(this.isChanged)
            {
                return;
            }

            this.isChanged = true;

            this.brightnessNumericUpDown.Value = this.brightnessTrackBar.Value;

            SetTargetColorTable();

            this.isChanged = false;
        }

        #endregion
        #region 명도 숫자 UP/DOWN 값 변경시 처리하기 - brightnessNumericUpDown_ValueChanged(sender, e)

        /// <summary>
        /// 명도 숫자 UP/DOWN 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void brightnessNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            if(this.isChanged)
            {
                return;
            }

            this.isChanged = true;

            this.brightnessTrackBar.Value = Convert.ToInt32(this.brightnessNumericUpDown.Value);

            SetTargetColorTable();

            this.isChanged = false;
        }

        #endregion
        #region 대비 트랙바 값 변경시 처리하기 - contrastTrackBar_ValueChanged(sender, e)

        /// <summary>
        /// 대비 트랙바 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void contrastTrackBar_ValueChanged(object sender, EventArgs e)
        {
            if(this.isChanged)
            {
                return;
            }

            this.isChanged = true;

            this.contrastNumericUpDown.Value = this.contrastTrackBar.Value;

            SetTargetColorTable();

            this.isChanged = false;
        }

        #endregion
        #region 대비 숫자 UP/DOWN 값 변경시 처리하기 - contrastNumericUpDown_ValueChanged(sender, e)

        /// <summary>
        /// 대비 숫자 UP/DOWN 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void contrastNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            if(this.isChanged)
            {
                return;
            }

            this.isChanged = true;

            this.contrastTrackBar.Value = Convert.ToInt32(this.contrastNumericUpDown.Value);

            SetTargetColorTable();

            this.isChanged = false;
        }

        #endregion
        #region 색상 트랙바 값 변경시 처리하기 - colorTrackBar_ValueChanged(sender, e)

        /// <summary>
        /// 색상 트랙바 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void colorTrackBar_ValueChanged(object sender, EventArgs e)
        {
            if(this.isChanged)
            {
                return;
            }

            this.isChanged = true;

            this.colorNumericUpDown.Value = this.colorTrackBar.Value;

            SetTargetColorTable();

            this.isChanged = false;
        }

        #endregion
        #region 색상 숫자 UP/DOWN 값 변경시 처리하기 - colorNumericUpDown_ValueChanged(sender, e)

        /// <summary>
        /// 색상 숫자 UP/DOWN 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void colorNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            if(this.isChanged)
            {
                return;
            }

            this.isChanged = true;

            this.colorTrackBar.Value = Convert.ToInt32(this.colorNumericUpDown.Value);

            SetTargetColorTable();

            this.isChanged = false;
        }

        #endregion

        #region 적용 버튼 클릭시 처리하기 - applyButton_Click(sender, e)

        /// <summary>
        /// 적용 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void applyButton_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.OK;
        }

        #endregion
        #region 취소 버튼 클릭시 처리하기 - cancelButton_Click(sender, e)

        /// <summary>
        /// 취소 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void cancelButton_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.Cancel;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 타겟 색상 테이블 설정하기 - SetTargetColorTable()

        /// <summary>
        /// 타겟 색상 테이블 설정하기
        /// </summary>
        private void SetTargetColorTable()
        {
            if(this.sourceColorTableControl.Value == null)
            {
                return;
            }

            List<Color> colorList = new List<Color>();

            foreach(Color color in this.sourceColorTableControl.Value)
            {
                double red   = color.R;
                double green = color.G;
                double blue  = color.B;

                double v, h, m, u;

                v = this.brightnessTrackBar.Value;

                if(v > 0)
                {
                    red   = red   + ((255d - red  ) * v / 100d);
                    green = green + ((255d - green) * v / 100d);
                    blue  = blue  + ((255d - blue ) * v / 100d);
                }
                else if(v < 0)
                {
                    red   *= (100d + v) / 100d;
                    green *= (100d + v) / 100d;
                    blue  *= (100d + v) / 100d;
                }

                v = (contrastTrackBar.Value + 100d) / 100d;
                h = (red + green + blue) / 3d;

                red   = Math.Max(0, Math.Min(255, h + (red   - h) * v));
                green = Math.Max(0, Math.Min(255, h + (green - h) * v));
                blue  = Math.Max(0, Math.Min(255, h + (blue  - h) * v));

                v = (colorTrackBar.Value + 360d) % 360d;
                m = (v % 120d) / 120d;
                u = (120d - (v % 120d)) / 120d;

                double nr, ng, nb;

                if(v < 120d)
                {
                    nr = (red   * u) + (green * m);
                    ng = (green * u) + (blue  * m);
                    nb = (blue  * u) + (red   * m);
                }
                else if(v < 240d)
                {
                    nr = (green * u) + (blue  * m);
                    ng = (blue  * u) + (red   * m);
                    nb = (red   * u) + (green * m);
                }
                else
                {
                    nr = (blue  * u) + (red   * m);
                    ng = (red   * u) + (green * m);
                    nb = (green * u) + (blue  * m);
                }

                red   = nr;
                green = ng;
                blue  = nb;

                byte redByte   = Convert.ToByte(Math.Max(0, Math.Min(255, red  )));
                byte greenByte = Convert.ToByte(Math.Max(0, Math.Min(255, green)));
                byte blueByte  = Convert.ToByte(Math.Max(0, Math.Min(255, blue )));

                colorList.Add(Color.FromArgb(redByte, greenByte, blueByte));
            }

            this.targetColorTableControl.Value = colorList;
        }

        #endregion
    }
}