﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// GIF 파트 헤더 컨트롤
    /// </summary>
    public partial class GIFPartHeaderControl : UserControl, IGIFControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 값 변경시 이벤트 - ValueChanged

        /// <summary>
        /// 값 변경시 이벤트
        /// </summary>
        public event EventHandler ValueChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 파트
        /// </summary>
        private GIFPartHeader part;

        /// <summary>
        /// 체크 여부
        /// </summary>
        private bool isChecked = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 파트 - Part

        /// <summary>
        /// 파트
        /// </summary>
        public GIFPartHeader Part
        {
            get
            {
                return this.part;
            }
            set
            {
                this.isChecked = false;

                this.part = value;

                SuspendLayout();

                if(value == null)
                {
                    this.versionComboBox.SelectedIndex       = 0;
                    this.widthNumericUpDown.Value            = 0;
                    this.heightNumericUpDown.Value           = 0;
                    this.colorResolutionNumericUpDown.Value  = 0;
                    this.backgroundNumericUpDown.Value       = 0;
                    this.aspectRatioNumericUpDown.Value      = 0;
                    this.sortColorTableCheckBox.Checked      = false;
                    this.useGlobalColorTableCheckBox.Checked = false;
                    this.colorTableSizeNumericUpDown.Value   = 0;
                    this.colorTableControl.Value             = new List<Color>();
                }
                else
                {
                    this.versionComboBox.SelectedIndex       = value.Version.Equals("89a") ? 0 : 1;
                    this.widthNumericUpDown.Value            = value.Width;
                    this.heightNumericUpDown.Value           = value.Height;
                    this.colorResolutionNumericUpDown.Value  = value.ColorResolution;
                    this.backgroundNumericUpDown.Value       = value.BackgroundColor;
                    this.aspectRatioNumericUpDown.Value      = value.ColorResolution;
                    this.sortColorTableCheckBox.Checked      = value.TableSorted;
                    this.useGlobalColorTableCheckBox.Checked = value.UseGlobalColorTable;
                    this.colorTableSizeNumericUpDown.Value   = value.GlobalColorTableSize;
                    this.colorTableControl.Value             = value.GlobalColorTable ?? new List<Color>();
                }

                ResumeLayout();

                this.isChecked = true;
            }
        }

        #endregion
        #region GIF 파트 - GIFPart

        /// <summary>
        /// GIF 파트
        /// </summary>
        public GIFPart GIFPart
        {
            get
            {
                return this.part;
            }
            set
            {
                Part = value as GIFPartHeader;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFPartHeaderControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public GIFPartHeaderControl()
        {
            InitializeComponent();

            this.versionComboBox.SelectedIndexChanged += (sender, e) => {

                if(this.part == null)
                {
                    return;
                }

                if(!this.isChecked)
                {
                    return;
                }

                try
                {
                    this.part.Version = versionComboBox.SelectedIndex == 0 ? "89a" : "87a";

                    FireValueChangedEvent();
                }
                catch
                {
                }
            };

            this.widthNumericUpDown.ValueChanged           += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.Width  = v                         ); };
            this.heightNumericUpDown.ValueChanged          += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.Height = v                         ); };
            this.colorResolutionNumericUpDown.ValueChanged += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.ColorResolution = Convert.ToByte(v)); };
            this.backgroundNumericUpDown.ValueChanged      += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.BackgroundColor = Convert.ToByte(v)); };
            this.aspectRatioNumericUpDown.ValueChanged     += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.AspectRatio = Convert.ToByte(v)    ); };
            this.sortColorTableCheckBox.CheckedChanged     += (sender, e) => { FireValueChangedEvent(sender as CheckBox     , v => this.part.TableSorted = v                    ); };
            this.colorTableControl.ValueChanged            += (sender, e) => {
            
                if(this.part != null && this.isChecked)
                {
                    FireValueChangedEvent();
                }
            };

            this.transformColorButton.Click += transformColorButton_Click;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 색상 변환 버튼 클릭시 처리하기 - transformColorButton_Click(sender, e)

        /// <summary>
        /// 색상 변환 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void transformColorButton_Click(object sender, EventArgs e)
        {
            if(this.colorTableControl.Value == null || this.colorTableControl.Value.Count == 0)
            {
                return;
            }

            GIFColorTableTransformForm form = new GIFColorTableTransformForm(this.colorTableControl.Value);

            if(form.ShowDialog() == DialogResult.OK)
            {
                for(int i = 0; i < form.Value.Count; i++)
                {
                    this.colorTableControl.Value[i] = form.Value[i];
                }

                this.colorTableControl.Value = this.colorTableControl.Value;

                FireValueChangedEvent();
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent()

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        private void FireValueChangedEvent()
        {
            if(ValueChanged != null)
            {
                try
                {
                    ValueChanged(this, EventArgs.Empty);
                }
                catch
                {
                }
            }
        }

        #endregion
        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent(numericUpDown, action)

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="numericUpDown">숫자 UP/DOWN</param>
        /// <param name="action">액션</param>
        private void FireValueChangedEvent(NumericUpDown numericUpDown, Action<ushort> action)
        {
            if(this.part == null)
            {
                return;
            }

            if(!this.isChecked)
            {
                return;
            }

            try
            {
                ushort value = Convert.ToUInt16(numericUpDown.Value);

                action(value);

                FireValueChangedEvent();
            }
            catch
            {
            }
        }

        #endregion
        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent(checkBox, action)

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="checkBox">체크 박스</param>
        /// <param name="action">액션</param>
        private void FireValueChangedEvent(CheckBox checkBox, Action<bool> action)
        {
            if(this.part == null)
            {
                return;
            }

            if(!this.isChecked)
            {
                return;
            }

            try
            {
                bool value = checkBox.Checked;

                action(value);

                FireValueChangedEvent();
            }
            catch
            {
            }
        }

        #endregion
    }
}