﻿using System;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// GIF 파트 메타 애플리케이션 컨트롤
    /// </summary>
    public partial class GIFPartMetaApplicationControl : UserControl, IGIFControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 값 변경시 이벤트 - ValueChanged

        /// <summary>
        /// 값 변경시 이벤트
        /// </summary>
        public event EventHandler ValueChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 파트
        /// </summary>
        private GIFPartMetaData.ApplicationData part;

        /// <summary>
        /// 체크 여부
        /// </summary>
        private bool isChecked = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 파트 - Part

        /// <summary>
        /// 파트
        /// </summary>
        public GIFPartMetaData.ApplicationData Part
        {
            get
            {
                return this.part;
            }
            set
            {
                this.isChecked = false;

                this.part = value;

                if(value == null)
                {
                    this.appIdentifierTextBox.Text = "";
                    this.hexadecimalControl.Value  = null;
                }
                else
                {
                    this.appIdentifierTextBox.Text = value.AppIdentifier;
                    this.hexadecimalControl.Value  = value.ContentByteArray;
                }

                this.isChecked = true;
            }
        }

        #endregion
        #region GIF 파트 - GIFPart

        /// <summary>
        /// GIF 파트
        /// </summary>
        public GIFPart GIFPart
        {
            get
            {
                return this.part;
            }
            set
            {
                Part = value as GIFPartMetaData.ApplicationData;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFPartMetaApplicationControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public GIFPartMetaApplicationControl()
        {
            InitializeComponent();

            this.appIdentifierTextBox.TextChanged += (sender, e) => {

                if(this.part == null || !this.isChecked)
                {
                    return;
                }

                try
                {
                    this.part.AppIdentifier = this.appIdentifierTextBox.Text;

                    FireValueChangedEvent();
                }
                catch
                {
                }
            };
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent()

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        private void FireValueChangedEvent()
        {
            if(ValueChanged != null)
            {
                try
                {
                    ValueChanged(this, EventArgs.Empty);
                }
                catch
                {
                }
            }
        }

        #endregion
    }
}