﻿using System;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// GIF 파트 메타 그래픽스 컨트롤 컨트롤
    /// </summary>
    public partial class GIFPartMetaGraphicsControl : UserControl, IGIFControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 값 변경시 이벤트 - ValueChanged

        /// <summary>
        /// 값 변경시 이벤트
        /// </summary>
        public event EventHandler ValueChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 파트
        /// </summary>
        private GIFPartMetaData.GraphicsControl part;

        /// <summary>
        /// 체크 여부
        /// </summary>
        private bool isChecked = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 파트 - Part

        /// <summary>
        /// 파트
        /// </summary>
        public GIFPartMetaData.GraphicsControl Part
        {
            get
            {
                return this.part;
            }
            set
            {
                this.isChecked = false;

                this.part = value;

                if(value == null)
                {
                    this.disposalMethodComboBox.SelectedIndex = 0;
                    this.useUserInputCheckBox.Checked         = false;
                    this.useTransparencyCheckBox.Checked      = false;
                    this.delayTimeNumericUpDown.Value         = 0;
                    this.transparentColorNumericUpDown.Value  = 0;
                }
                else
                {
                    this.disposalMethodComboBox.SelectedIndex = (int)value.DisposalMethod;
                    this.useUserInputCheckBox.Checked         = value.UseUserInput;
                    this.useTransparencyCheckBox.Checked      = value.UseTransparency;
                    this.delayTimeNumericUpDown.Value         = value.DelayTime;
                    this.transparentColorNumericUpDown.Value  = value.TransparentColor;
                }

                this.isChecked = true;
            }
        }

        #endregion
        #region GIF 파트 - GIFPart

        /// <summary>
        /// GIF 파트
        /// </summary>
        public GIFPart GIFPart
        {
            get
            {
                return this.part;
            }
            set
            {
                Part = value as GIFPartMetaData.GraphicsControl;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFPartMetaGraphicsControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public GIFPartMetaGraphicsControl()
        {
            InitializeComponent();

            this.disposalMethodComboBox.SelectedIndex = 0;

            this.disposalMethodComboBox.SelectedIndexChanged += (sender, e) => {

                if(this.part == null || !this.isChecked)
                {
                    return;
                }

                try
                {
                    this.part.DisposalMethod = (GIFPartMetaData.DisposalMethod)this.disposalMethodComboBox.SelectedIndex;

                    FireValueChangedEvent();
                }
                catch
                {
                }
            };

            this.useUserInputCheckBox.CheckedChanged        += (sender, e) => { FireValueChangedEvent(sender as CheckBox     , v => this.part.UseUserInput = v                    ); };
            this.useTransparencyCheckBox.CheckedChanged     += (sender, e) => { FireValueChangedEvent(sender as CheckBox     , v => this.part.UseTransparency = v                 ); };
            this.delayTimeNumericUpDown.ValueChanged        += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.DelayTime = v                       ); };
            this.transparentColorNumericUpDown.ValueChanged += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.TransparentColor = Convert.ToByte(v)); };
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent()

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        private void FireValueChangedEvent()
        {
            if(ValueChanged != null)
            {
                try
                {
                    ValueChanged(this, EventArgs.Empty);
                }
                catch
                {
                }
            }
        }

        #endregion
        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent(numericUpDown, action)

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="numericUpDown">숫자 UP/DOWN</param>
        /// <param name="action">액션</param>
        private void FireValueChangedEvent(NumericUpDown numericUpDown, Action<ushort> action)
        {
            if(this.part == null)
            {
                return;
            }

            if(!this.isChecked)
            {
                return;
            }

            try
            {
                ushort value = Convert.ToUInt16(numericUpDown.Value);

                action(value);

                FireValueChangedEvent();
            }
            catch
            {
            }
        }

        #endregion
        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent(checkBox, action)

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="checkBox">체크 박스</param>
        /// <param name="action">액션</param>
        private void FireValueChangedEvent(CheckBox checkBox, Action<bool> action)
        {
            if(this.part == null)
            {
                return;
            }

            if(!this.isChecked)
            {
                return;
            }

            try
            {
                bool value = checkBox.Checked;

                action(value);

                FireValueChangedEvent();
            }
            catch
            {
            }
        }

        #endregion
    }
}