﻿using System;
using System.Windows.Forms;

namespace TestProject
{
    /// <summary>
    /// GIF 파트 메타 텍스트 컨트롤
    /// </summary>
    public partial class GIFPartMetaTextControl : UserControl, IGIFControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 값 변경시 이벤트 - ValueChanged

        /// <summary>
        /// 값 변경시 이벤트
        /// </summary>
        public event EventHandler ValueChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 파트
        /// </summary>
        private GIFPartMetaData.TextDraw part;

        /// <summary>
        /// 체크 여부
        /// </summary>
        private bool isChecked = false;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 파트 - Part

        /// <summary>
        /// 파트
        /// </summary>
        public GIFPartMetaData.TextDraw Part
        {
            get
            {
                return this.part;
            }
            set
            {
                this.isChecked = false;

                this.part = value;

                this.leftNumericUpDown.Value            = value.Left;
                this.topNumericUpDown.Value             = value.Top;
                this.widthNumericUpDown.Value           = value.Width;
                this.heightNumericUpDown.Value          = value.Height;
                this.characterWidthNumericUpDown.Value  = value.CharacterWidth;
                this.characterHeightNumericUpDown.Value = value.CharacterHeight;
                this.textColorNumericUpDown.Value       = value.TextColor;
                this.backgroundColorNumericUpDown.Value = value.BackgroundColor;
                this.textTextBox.Text                   = value.Text;

                this.isChecked = true;
            }
        }

        #endregion
        #region GIF 파트 - GIFPart

        /// <summary>
        /// GIF 파트
        /// </summary>
        public GIFPart GIFPart
        {
            get
            {
                return this.part;
            }
            set
            {
                Part = value as GIFPartMetaData.TextDraw;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - GIFPartMetaTextControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public GIFPartMetaTextControl()
        {
            InitializeComponent();

            this.leftNumericUpDown.ValueChanged            += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.Left            = v                ); };
            this.topNumericUpDown.ValueChanged             += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.Top             = v                ); };
            this.widthNumericUpDown.ValueChanged           += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.Width           = v                ); };
            this.heightNumericUpDown.ValueChanged          += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.Height          = v                ); };
            this.characterWidthNumericUpDown.ValueChanged  += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.CharacterWidth  = Convert.ToByte(v)); };
            this.characterHeightNumericUpDown.ValueChanged += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.CharacterHeight = Convert.ToByte(v)); };
            this.textColorNumericUpDown.ValueChanged       += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.TextColor       = Convert.ToByte(v)); };
            this.backgroundColorNumericUpDown.ValueChanged += (sender, e) => { FireValueChangedEvent(sender as NumericUpDown, v => this.part.BackgroundColor = Convert.ToByte(v)); };
            this.textTextBox.TextChanged                   += (sender, e) => {

                if(this.part == null || !this.isChecked)
                {
                    return;
                }

                try
                {
                    this.part.Text = textTextBox.Text;

                    FireValueChangedEvent();
                }
                catch
                {
                }
            };
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent()

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        private void FireValueChangedEvent()
        {
            if(ValueChanged != null)
            {
                try
                {
                    ValueChanged(this, EventArgs.Empty);
                }
                catch
                {
                }
            }
        }

        #endregion
        #region 값 변경시 이벤트 발생시키기 - FireValueChangedEvent(numericUpDown, action)

        /// <summary>
        /// 값 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="numericUpDown">숫자 UP/DOWN</param>
        /// <param name="action">액션</param>
        private void FireValueChangedEvent(NumericUpDown numericUpDown, Action<ushort> action)
        {
            if(this.part == null)
            {
                return;
            }

            if(!this.isChecked)
            {
                return;
            }

            try
            {
                ushort value = Convert.ToUInt16(numericUpDown.Value);

                action(value);

                FireValueChangedEvent();
            }
            catch
            {
            }
        }

        #endregion
    }
}